use strict;
use warnings;
use autodie;

use Test::More tests => 269;
use Stanford::LDAP::SyncSuGAL;
use Net::LDAPapi;
use Data::Compare;
use Data::Dumper;

require 't/setup.pl';

# Check for command-line option overrides (not in build mode)
my $S;
my $tests_to_run;
my $WAIT_TIME;
if (!$ENV{'DEB_PACKAGE_BUILD'}) {
    use Getopt::Long::Descriptive;
    my ($OPTIONS, $usage) = describe_options(
       '%c <action> %o',
        [ 'krb5ccname_path=s', 'use this ticket file for credentials', ],
        [ 'ldap_host=s',       'test against this LDAP master server', ],
        [ 'tests=s',           'which tests to run', ],
        [ 'wait_time=i',       'give listener this time to finish (in seconds)', ],
       );

    $S = do_setup(undef, $OPTIONS->ldap_host, $OPTIONS->krb5ccname_path);
    $tests_to_run = $OPTIONS->tests;
    if ($OPTIONS->wait_time()) {
       $WAIT_TIME = $OPTIONS->wait_time();
    } else {
        $WAIT_TIME = 5;
    }

} else {
    $S = do_setup();
}

my %VISIB_ATTRIBUTES = %{ $S->VISIB_ATTRIBUTES() };

my $user_dn = 'suRegID=111823c0c3fe409e9fbde77fee52b100,cn=people,dc=stanford,dc=edu';

my $USE_LISTENER = 1 ;


# #### #### #### #### #### #### #### #### #### #### #### #### #### ####
sub do_sync {
    my ($wait_time) = @_ ;

    if ($USE_LISTENER) {
        if (!defined($wait_time)) {
            $wait_time = $WAIT_TIME;
        }
        diag("waiting $wait_time seconds for the listener to catch up...");
        sleep $wait_time;
        diag("...done waiting");
    } else {
        die "????";
    }
    return;
}
# #### #### #### #### #### #### #### #### #### #### #### #### #### ####
sub random_text {
    my ($length) = @_ ;
    my @chars = ("A".."Z", "a".."z");
    my $string = q{};
    $string .= $chars[rand @chars] for 1..$length;
    return $string;
}
# #### #### #### #### #### #### #### #### #### #### #### #### #### ####


# TEST SET 1.
# Make several changes, and then wait for the listener to catch-up
#
#  * set suVisibIdentity visibility to 'stanford'
#  * set sn to Smithee
#  * set sn to Roberts
#  * set sn to Jonesey
sub test01_basic {
    diag("Starting TEST SET 1 (basic)");

    ok($S->set_attribute($user_dn, 'suVisibIdentity', ['stanford']));
    ok($S->get_attribute($user_dn, 'suVisibIdentity') eq 'stanford');

    ok($S->set_attribute($user_dn, 'sn', ['Smithee']));
    ok($S->get_attribute($user_dn, 'sn') eq 'Smithee');

    ok($S->set_attribute($user_dn, 'sn', ['Roberts']));
    ok($S->get_attribute($user_dn, 'sn') eq 'Roberts');

    ok($S->set_attribute($user_dn, 'sn', ['Lewenberg']));
    ok($S->get_attribute($user_dn, 'sn') eq 'Lewenberg');

    do_sync();

    ok($S->get_attribute($user_dn, 'suGALsn') eq 'Lewenberg');
    diag("End of TEST SET 1");
}

# TEST SET 2.
# Make the visibility private and verify that the suGAL settings for that
# visibility control go away. Then set it back and make sure the suGAL
# settings come back.
#
#  * set suVisibIdentity visibility to 'private'
sub test02_privacy {
    diag("Starting TEST SET 2 (setting to private removes suGAL)");

    ok($S->set_attribute($user_dn, 'suVisibIdentity', ['private']));
    ok($S->get_attribute($user_dn, 'suVisibIdentity') eq 'private');

    do_sync();

    # Verify that ALL the suGAL attributes are UNDEFINED.
    my %suVisibIdentity_attributes = %{ $VISIB_ATTRIBUTES{'suVisibIdentity'} };
    my @suGALattrs = values %suVisibIdentity_attributes ;
    foreach my $suGALattr (@suGALattrs) {
        ok(!defined($S->get_attribute($user_dn, $suGALattr)), $suGALattr);
    }

    ok($S->set_attribute($user_dn, 'suVisibIdentity', ['stanford']));
    ok($S->get_attribute($user_dn, 'suVisibIdentity') eq 'stanford');

    do_sync();
    ok($S->get_attribute($user_dn, 'suGALsn') eq 'Lewenberg');

    diag("End of TEST SET 2");
}


# TEST SET 3.
# Set one of the multi-valued attributes and make sure the suGAL
# attributes are set properly.
#
#  * set suSearchID to ['adamhl', 'adamlewenberg']
#  * check that suGALSearchID is set correctly.
#  * set suSearchID to ['adamhl']
#  * check that suGALSearchID is set correctly.
#  * set suSearchID to []
#  * check that suGALSearchID is set correctly.
sub test03_multi_valued {
    diag("Starting TEST SET 3 (multi-valued)");

    # Make sure suVisibSunetID is not set to 'private'
    ok($S->set_attribute($user_dn, 'suVisibSunetID', ['stanford']));
    ok($S->get_attribute($user_dn, 'suVisibSunetID') eq 'stanford');

    my (@suSearchIDs, @suGALsuSearchIDs);
    ok($S->set_attribute($user_dn, 'suSearchID', ['adamhl', 'adamlewenberg']));
    @suSearchIDs = $S->get_attribute($user_dn, 'suSearchID');
    ok (Compare(\@suSearchIDs, ['adamhl', 'adamlewenberg'] ));

    do_sync();

    @suGALsuSearchIDs = $S->get_attribute($user_dn, 'suGALsuSearchID');
    @suGALsuSearchIDs = sort @suGALsuSearchIDs;
    ok (Compare(\@suGALsuSearchIDs, ['adamhl', 'adamlewenberg'] ))
      or diag('@suGALsuSearchIDs is [' . join(q{,}, @suGALsuSearchIDs)
               . q{]; should have been [adamhl, adamlewenberg]});

    ok($S->set_attribute($user_dn, 'suSearchID', ['adamhl']));
    @suSearchIDs = $S->get_attribute($user_dn, 'suSearchID');
    ok (Compare(\@suSearchIDs, ['adamhl'] ));

    do_sync();

    @suGALsuSearchIDs = $S->get_attribute($user_dn, 'suGALsuSearchID');
    ok (Compare(\@suGALsuSearchIDs, ['adamhl'] ));

    # Remove all the suSearchID attributes.
    ok($S->set_attribute($user_dn, 'suSearchID', []));
    @suSearchIDs = $S->get_attribute($user_dn, 'suSearchID');
    ok (Compare(\@suSearchIDs, [] ));

    do_sync();

    @suGALsuSearchIDs = $S->get_attribute($user_dn, 'suGALsuSearchID');
    ok (Compare(\@suGALsuSearchIDs, [] ));

    # Set back to original.
    ok($S->set_attribute($user_dn, 'suSearchID', ['adamhl', 'adamlewenberg']));
    @suSearchIDs = $S->get_attribute($user_dn, 'suSearchID');
    ok (Compare(\@suSearchIDs, ['adamhl', 'adamlewenberg'] ));

    do_sync();

    @suGALsuSearchIDs = $S->get_attribute($user_dn, 'suGALsuSearchID');
    @suGALsuSearchIDs = sort @suGALsuSearchIDs;
    ok (Compare(\@suGALsuSearchIDs, ['adamhl', 'adamlewenberg'] ))
      or diag('@suGALsuSearchIDs is [' . join(q{, }, @suGALsuSearchIDs)
              . q{]; should have been [adamhl, adamlewenberg]});


    diag("End of TEST SET 3");
}

# Set (most) all the suVisibIdentity-controlled attributes and verify that all
# the suGAL attributes get set as well.
sub test_all_attributes {
    my ($counter, $visibility_attribute, $person_attributes_aref) = @_ ;
    diag("Starting TEST SET $counter ($visibility_attribute)");

    my @current_attributes;

    my @saved_visibility_attributes = $S->get_attribute($user_dn, $visibility_attribute);
    ok($S->set_attribute($user_dn, $visibility_attribute, ['stanford']));
    @current_attributes = $S->get_attribute($user_dn, $visibility_attribute);
    ok(Compare(\@current_attributes, ['stanford']));


    my $debug = 0 ;
    my $progress = sub {
        if (!$debug) { return ; }
        my ($msg) = @_ ;
        print $msg . "\n" ;
    };

    my @person_attrs = @$person_attributes_aref;

    my %attr_to_gal = %{ $VISIB_ATTRIBUTES{$visibility_attribute} };
    my @suGAL_attrs =  values %attr_to_gal;

    # Save the attributes.
    my %person_attr_to_values = ();
    foreach my $person_attr (@person_attrs) {
        my @values = $S->get_attribute($user_dn, $person_attr);
        if (scalar(@values == 0)) {
            $person_attr_to_values{$person_attr} = undef;
        } else {
            $person_attr_to_values{$person_attr} = \@values;
        }
    }

    # Set some new ones.
    my %person_attr_to_new_values = ();
    foreach my $person_attr (@person_attrs) {
        my $new_value = random_text(10);
        $person_attr_to_new_values{$person_attr} = $new_value;
        ok($S->set_attribute($user_dn, $person_attr, [$new_value]),
            "setting attribute $person_attr to value '$new_value'");
    }

    do_sync();

    # Check that the suGAL are set correctly.
    foreach my $person_attr (@person_attrs) {
        my $suGALattr   = $attr_to_gal{$person_attr};
        my @suGALvalues = $S->get_attribute($user_dn, $suGALattr);
        $progress->("suGALattr is $suGALattr") ;
        $progress->("suGALvalues are " . join(q{,}, @suGALvalues)) ;
        ok (Compare(\@suGALvalues, [$person_attr_to_new_values{$person_attr}]),
            $person_attr);
    }

    # Reset the person values to their original values.
    foreach my $person_attr (@person_attrs) {
        my $orig_values_aref = $person_attr_to_values{$person_attr};
        if (defined($orig_values_aref)) {
            ok($S->set_attribute($user_dn, $person_attr,
                                 $orig_values_aref));
        }
    }

    foreach my $person_attr (@person_attrs) {
        my $orig_values_aref = $person_attr_to_values{$person_attr};
        if (defined($orig_values_aref)) {
            my @values = $S->get_attribute($user_dn, $person_attr);
            ok (Compare(\@values, $person_attr_to_values{$person_attr}));
        }
    }

    @saved_visibility_attributes = $S->get_attribute($user_dn, $visibility_attribute);
    $progress->('saved visibility attribute ' . join(q{,}, @saved_visibility_attributes));

    ok($S->set_attribute($user_dn, $visibility_attribute, \@saved_visibility_attributes));

    @current_attributes = $S->get_attribute($user_dn, $visibility_attribute);
    $progress->('current visibility attribute ' . join(q{,}, @current_attributes));

    ok(Compare(\@current_attributes, \@saved_visibility_attributes));

    diag("End of TEST SET $counter");
}

sub test04_all_identity {
    my @person_attrs = qw/
        displayName
        generationQualifier
        givenName
        personalTitle
        sn
        suDisplayNameFirst
        suDisplayNameLast
        suDisplayNameLF
        suDisplayNameMiddle
        suDisplayNamePrefix
        suDisplayNameSuffix
        suFacultyAppointment
        suFacultyAppointmentShort
        suOtherName
        suUniqueIdentifier
        /;

    return test_all_attributes(4, 'suVisibIdentity', \@person_attrs);
}

sub test05_all_email_pager {
    my @person_attrs = qw/
        suEmailPager
        /;

    return test_all_attributes(5, 'suVisibPagerEmail', \@person_attrs);
}

# Because facsimileTelephoneNumber and suGALfacsimileTelephoneNumber do
# not have EQUALITY rules, updating these fields requires a complete
# delete and then an add. This tests that functionality.
sub test06_all_facsimile {

    # Make sure we can set and change facsimileTelephoneNumber
    ok($S->set_attribute($user_dn, 'facsimileTelephoneNumber', ['650-555-5555']));
    ok($S->get_attribute($user_dn, 'facsimileTelephoneNumber') eq '650-555-5555');

    ok($S->set_attribute($user_dn, 'facsimileTelephoneNumber', ['650-555-6666']));
    ok($S->get_attribute($user_dn, 'facsimileTelephoneNumber') eq '650-555-6666');

    ok($S->set_attribute($user_dn, 'facsimileTelephoneNumber', ['650-555-6666', '650-555-7777']));
    my @faxes = $S->get_attribute($user_dn, 'facsimileTelephoneNumber');

    ok (Compare(\@faxes, ['650-555-6666', '650-555-7777'] ));

    # Set to private and make sure that all the
    # suGALfacsimileTelephoneNumber attributes get deleted.
    ok($S->set_attribute($user_dn, 'suVisibFacsimileTelephoneNumber', ['private']));
    ok($S->get_attribute($user_dn, 'suVisibFacsimileTelephoneNumber') eq 'private');
    do_sync();

    @faxes = $S->get_attribute($user_dn, 'suGALfacsimileTelephoneNumber');
    ok (Compare(\@faxes, []), 'faxes should be empty');

    # Set visibility back to 'stanford'.
    ok($S->set_attribute($user_dn, 'suVisibFacsimileTelephoneNumber', ['stanford']));
    do_sync();

    @faxes = $S->get_attribute($user_dn, 'suGALfacsimileTelephoneNumber');
    ok (Compare(\@faxes, ['650-555-6666', '650-555-7777']),
        'suGALfacsimileTelephoneNumber has both fax numbers');

    # Now test all the attributes
    my @person_attrs = qw/
        facsimileTelephoneNumber
        /;

    return test_all_attributes(6, 'suVisibFacsimileTelephoneNumber', \@person_attrs);
}

sub test07_all_home_address {
    # We leave out 'c' since it has a special format.
    my @person_attrs = qw/
        homePhone
        homePostalAddress
        l
        postalAddress
        postalCode
        st
        street
        /;

    return test_all_attributes(7, 'suVisibHomeAddress', \@person_attrs);
}

sub test08_all_home_page {
    my @person_attrs = qw/
        labeledURI
        /;

    return test_all_attributes(8, 'suVisibHomePage', \@person_attrs);
}

sub test09_all_email {
    my @person_attrs = qw/
        mail
        /;

    return test_all_attributes(9, 'suVisibEmail', \@person_attrs);
}

sub test10_all_mobile_phone {
    my @person_attrs = qw/
        mobile
        /;

    return test_all_attributes(10, 'suVisibMobilePhone', \@person_attrs);
}

sub test11_all_pager_phone {
    my @person_attrs = qw/
        pager
        /;

    return test_all_attributes(11, 'suVisibPagerPhone', \@person_attrs);
}

sub test12_all_local_address {
    my @person_attrs = qw/
       suLocalAddress
       suResidenceRoom
        /;

    # Need to add suCampusResident objectClass first.
    my $cnt = $S->get_entry_cnt($user_dn, 'objectClass=suCampusResident', LDAP_SCOPE_BASE);
    if (!$cnt) {
         my %attributes = (
             "objectClass", {'a', ["suCampusResident"]},
             "suResidenceRequiredAttribute", {'a', ["ZZZ"]},
             );
         $S->LDAP_UPDATE()->modify_s($user_dn, \%attributes);
    }

    return test_all_attributes(12, 'suVisibLocalAddress', \@person_attrs);
}

sub test13_all_local_phone {
    my @person_attrs = qw/
       suLocalPhone
       suResidencePhone
        /;

    return test_all_attributes(13, 'suVisibLocalPhone', \@person_attrs);
}

sub test14_all_mail_address {
    my @person_attrs = qw/
     suMailAddress
        /;

    return test_all_attributes(14, 'suVisibMailAddress', \@person_attrs);
}

sub test15_all_permanent_address {
    my @person_attrs = qw/
     suPermanentAddress
        /;

    return test_all_attributes(15, 'suVisibPermanentAddress', \@person_attrs);
}

sub test16_all_permanent_phone {
    my @person_attrs = qw/
     suPermanentPhone
        /;

    return test_all_attributes(16, 'suVisibPermanentPhone', \@person_attrs);
}

sub test17_all_sunetid {
    my @person_attrs = qw/
      suSearchID
      suSunetID
        /;

    return test_all_attributes(17, 'suVisibSunetID', \@person_attrs);
}

sub test18_all_sunetid {
    my @person_attrs = qw/
      telephoneNumber
        /;

    return test_all_attributes(18, 'suVisibTelephoneNumber', \@person_attrs);
}

sub test19_multiple_susunetid {
    # This should fail (cannot have two suSunetID attributes that are
    # equal modulo case.
    ok(!$S->set_attribute($user_dn, 'suSunetID', ['alewenberg', 'ALewenberg']));
}

SKIP: {
    skip 'building the package', 248 if $ENV{'DEB_PACKAGE_BUILD'};

    my %test_names = (
         1 => \&test01_basic,
         2 => \&test02_privacy,
         3 => \&test03_multi_valued,
         4 => \&test04_all_identity,
         5 => \&test05_all_email_pager,
         6 => \&test06_all_facsimile,
         7 => \&test07_all_home_address,
         8 => \&test08_all_home_page,
         9 => \&test09_all_email,
        10 => \&test10_all_mobile_phone,
        11 => \&test11_all_pager_phone,
        12 => \&test12_all_local_address,
        13 => \&test13_all_local_phone,
        14 => \&test14_all_mail_address,
        15 => \&test15_all_permanent_address,
        16 => \&test16_all_permanent_phone,
        17 => \&test17_all_sunetid,
        18 => \&test18_all_sunetid,
        19 => \&test19_multiple_susunetid,
    );

    # What tests should we run?
    my @tests;
    if (!$tests_to_run || ($tests_to_run eq '0')) {
        @tests = 1..scalar(keys %test_names);
    } else {
        @tests = split(q{,}, $tests_to_run);
    }

    foreach my $test (@tests) {
        my $test_fref = $test_names{$test};
        $test_fref->();
    }
}

# clean-up
if (-f $S->tgt_file()) {
   unlink $S->tgt_file();
}
